#include "m_pd.h"
#include <stdlib.h>
#include <math.h>
#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

/* ------------------------ butter~ ----------------------------- */
/*LP and HP filter
 
 Type : biquad, tweaked butterworth
 References : Posted by Patrice Tarrabia
 Code :
 r = rez amount, from sqrt(2) to ~ 0.1
 f = cutoff frequency
 (from ~0 Hz to SampleRate/2 - though many
 synths seem to filter only up to SampleRate/4)
 
 The filter algo:
 out(n) = a1 * in + a2 * in(n-1) + a3 * in(n-2) - b1*out(n-1) - b2*out(n-2)
 
 Lowpass:
 c = 1.0 / tan(pi * f / sample_rate);
 
 a1 = 1.0 / ( 1.0 + r * c + c * c);
 a2 = 2* a1;
 a3 = a1;
 b1 = 2.0 * ( 1.0 - c*c) * a1;
 b2 = ( 1.0 - r * c + c * c) * a1;
 
 Hipass:
 c = tan(pi * f / sample_rate);
 
 a1 = 1.0 / ( 1.0 + r * c + c * c);
 a2 = -2*a1;
 a3 = a1;
 b1 = 2.0 * ( c*c - 1.0) * a1;
 b2 = ( 1.0 - r * c + c * c) * a1;
 */

static t_class *butter_class;

typedef struct _butter
{
    t_object x_obj; 	/* obligatory header */
    t_float frequency;    	/* floats on the first inlet set the frequency */
    t_float oldfrequency;
	t_float	quality;	/* the filter q or quality */
    t_float sample_rate;
    t_float c;
    t_float a1, a2, a3;
    t_float b1, b2;
    t_float x1, x2;
    t_float y1, y2;
} t_butter;

    /* this is the actual performance routine which acts on the samples.
    It's called with a single pointer "w" which is our location in the
    DSP call list.  We return a new "w" which will point to the next item
    after us.  Meanwhile, w[0] is just a pointer to dsp-perform itself
    (no use to us), w[1] and w[2] are the input and output vector locations,
    and w[3] is the number of points to calculate. */
static t_int *butter_perform(t_int *w)
{
	t_butter *x = (t_butter *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    t_float *out = (t_float *)(w[3]);
    int n = (int)(w[4]);

	// i like counting from zero, so i use sample to count the offset from
	// the start of the in and out blocks
	int sample = 0;
    float omega;

    if(x->frequency != x->oldfrequency && x->frequency > 0.0 && x->frequency < (x->sample_rate * 0.49))
    {
        x->oldfrequency = x->frequency;
        omega = x->frequency/x->sample_rate;
        if(omega < 0.1)
            x->c = 1.0 / (omega * 3.141592653589793);
        else
            x->c = tan((0.5 - omega) * 3.141592653589793);
     }
    x->a1 = 1.0 / ( 1.0 + x->quality * x->c + x->c * x->c);
    x->a2 = 2.0 * x->a1;
    x->a3 = x->a1;
    x->b1 = 2.0 * ( 1.0 - x->c * x->c) * x->a1;
    x->b2 = ( 1.0 - x->quality * x->c + x->c * x->c) * x->a1;

    while(n--)
    {
        *(out + sample) = x->a1 * *(in + sample) + x->a2 * x->x1 + x->a3 * x->x1 - x->b1 * x->y1 - x->b2 * x->y2;
        
        x->x2 = x->x1;
        x->x1 = *(in + sample);
        x->y2 = x->y1;
        x->y1 = *(out + sample);
		sample++;
    }
    return (w+5);
}

    /* called to start DSP.  Here we call Pd back to add our perform
    routine to a linear callback list which Pd in turn calls to grind
    out the samples. */
static void butter_dsp(t_butter *x, t_signal **sp)
{
    x->x1 = x->x2 = x->y1 = x->y2 = 0.0;
    x->oldfrequency = 10001.0;
    x->frequency = 10000.0;
    x->quality = 0.5;
	x->sample_rate = sp[0]->s_sr;
    dsp_add(butter_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}

void butter_quality(t_butter *x, t_floatarg g)
{
    if(g > 1.0)
        x->quality = 0.1;
    else if(g < 0.0)
        x->quality = 1.414;
    else
        x->quality = 0.1 + (1.0 - g) * 1.314;
}

static void *butter_new(void)
{
    t_butter *x = (t_butter *)pd_new(butter_class);
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("quality"));
    outlet_new(&x->x_obj, gensym("signal"));
    return (x);
}

// since we allocated some memory, we need a delete function
static void butter_free(t_butter *x)
{
}

    /* this routine, which must have exactly this name (with the "~" replaced
    by "_tilde) is called when the code is first loaded, and tells Pd how
    to build the "class". */
void butter_tilde_setup(void)
{
    butter_class = class_new(gensym("butter~"), (t_newmethod)butter_new, (t_method)butter_free,
    	sizeof(t_butter), 0, A_DEFFLOAT, 0);
	    /* this is magic to declare that the leftmost, "main" inlet
	    takes signals; other signal inlets are done differently... */
		/* also installs delay_time as the leftmost inlet float */
    CLASS_MAINSIGNALIN(butter_class, t_butter, frequency);
    	/* here we tell Pd about the "dsp" method, which is called back
	when DSP is turned on. */
    class_addmethod(butter_class, (t_method)butter_dsp, gensym("dsp"), (t_atomtype)0);
    class_addmethod(butter_class, (t_method)butter_quality, gensym("quality"), A_FLOAT, (t_atomtype)0);
}
